"""
Elliptic curves
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import sage.rings.arith as arith
import sage.misc.errors as errors
import sage.misc.misc as misc
import sage.modular.modform as modform
import sage.rings.rings as rings
import sage.tables.cremona as table
import sage.pari.wrap

oo = rings.infinity       # infinity
Q = rings.RationalField()         # rational numbers
O = rings.O         # big oh 

class EllipticCurve(object):
    def __init__(self, ainvs):
        self.__np = {}
        if isinstance(ainvs, str):
            #X = tables.Cremona.curve(ainvs)
            X = table.curve(ainvs)
            self.__ainvs = X.a_invariants
            self.__base_field = Q
            self.__analytic_rank = X.rank
            self.__rank = X.rank
            self.__conductor = X.conductor
            self.__torsion_order = X.torsion
            self.__cremona_label = "%s%s%s"%(X.conductor,X.iso_class,X.number)
            self.__tamagawa_number = X.tamagawa
            return
        if len(ainvs) == 2:
            ainvs = [0,0,0] + ainvs
        elif len(ainvs) != 5:
            raise errors.UsageError("Number of invariants must be 5.")
        if not isinstance(ainvs[0], rings.Element) or not ainvs[0].parent().is_field():
            try:
                ainvs = [Q(a) for a in ainvs]
            except:
                raise errors.MathError("Defining coefficients must lie in a field.")
        self.__base_field = ainvs[0].parent()
        self.__ainvs = [self.__base_field(x) for x in ainvs]

    def __repr__(self):
        a = self.__ainvs
        #return "y^2 + %s*x*y + %s*y = x^3 + %s*x^2 + %s*x + %s"%\
        #       (a[0], a[2], a[1], a[3], a[4])
        s = "y^2 "
        if a[0] == -1:
            s += "- x*y "
        elif a[0] == 1:
            s += "+ x*y "
        elif a[0] != 0:
            s += "%s*x*y "%a[0]
        if a[2] == -1:
            s += " - y"
        elif a[2] == 1:
            s += "+ y"
        elif a[2] != 0:
            s += "+ %s*y"%a[2]
        s += " = x^3 "
        if a[1] == -1:
            s += "- x^2 "
        elif a[1] == 1:
            s += "+ x^2 "
        elif a[1] != 0:
            s += "+ %s*x^2 "%a[1]
        if a[3] == -1:
            s += "- x "
        elif a[3] == 1:
            s += "+ x "
        elif a[3] != 0:
            s += "+ %s*x "%a[3]
        if a[4] == -1:
            s += "-1"
        elif a[4] == 1:
            s += "+1"
        elif a[4] != 0:
            s += "+ %s"%a[4]
        return s

    def __contains__(self, P):
        if not isinstance(P, Point):
            P = self(P)
        if P.curve() == self:
            return True
        x, y, a = P[0], P[1], self.__ainvs
        return y**2 + a[0]*x*y + a[2]*y == x**3 + a[1]*x**2 + a[3]*x + a[4]

    def __call__(self, x=oo, y=oo):
        if isinstance(x,Point):
            if x.curve() == self:
                return x
            else:
                x, y = x[0], x[1]
        elif isinstance(x,list):
            if len(x) != 2:
                raise errors.UsageError("List must have length 2.")
            x, y = x[0], x[1]
        return Point(self, x, y)

    def __getitem__(self, n):
        """
        """
        raise NotImplementedError, "not implemented."

    def ainvs(self):
        """
        The a-invariants of the elliptic curve.
        """
        return self.__ainvs

    def a_invariants(self):
        """
        The a-invariants of the elliptic curve.
        """
        return self.__ainvs

    def base_field(self):
        return self.__base_field

    def a1(self):
        return self.__ainvs[0]
    
    def a2(self):
        return self.__ainvs[1]
    
    def a3(self):
        return self.__ainvs[2]
    
    def a4(self):
        return self.__ainvs[3]
    
    def a6(self):
        return self.__ainvs[4]

    

    def Np(self, p):
        """
        The number of points on E modulo p, where p is a prime
        of good reduction.
        """
        return p+1-self.ap(p)
        
    def __get_pari_curve(self):
        try:
            return self.__pari_curve
        except AttributeError:
            self.__pari_curve = sage.pari.wrap.PariObject("ellinit(%s)"%self.__ainvs)
            return self.__pari_curve
        
    pari_curve = property(__get_pari_curve, \
            doc='PARI object underlying this elliptic curve')

    def __get_pari_mincurve(self):
        try:
            return self.__pari_mincurve
        except AttributeError:
            e = self.pari_curve
            R = e.ellglobalred()
            self.__conductor = int(R[0])
            self.__pari_mincurve = e.ellchangecurve(R[1])
            return self.__pari_mincurve

    pari_mincurve = property(__get_pari_mincurve, \
                             doc = "PARI elliptic curve in minimal " + \
                             "Weierstrass form isomorphic to this curve.")
    

    def anlist(self, n):
        """
        The first n Fourier coefficients of the modular form attached
        to this elliptic curve.  The ith element of the return list is
        a[i].
        

        EXAMPLES:
            >>> E=EllipticCurve([0,1])
            >>> E.anlist(20)
            [0, 1, 0, 0, 0, 0, 0, -4, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 8, 0]
        """
        try:
            if len(self.__anlist) > n:
                return self.__anlist[:n+1]
        except AttributeError:
            pass
        E = self.pari_mincurve
        self.__anlist = [0] + eval(str(E.ellan(n)))
        return self.__anlist

    def analytic_rank(self):
        try:
            return self.__analytic_rank
        except AttributeError:
            raise NotImplementedError
            #x = pari.ellanalyticrank(self)
            #self.__analytic_rank = x[0]
            return self.__analytic_rank

    def rank(self):
        try:
            return self.__rank
        except AttributeError:
            misc.todo("Implement rank.")
            
    def an(self, n):
        """
        The n-th Fourier coefficient of the modular form corresponding
        to this elliptic curve, where n is a positive integer.
        """
        return int(self.pari_mincurve.ellak(n))

    def ap(self, p):
        """
        The p-th Fourier coefficient of the modular form corresponding
        to this elliptic curve, where p is prime.
        """
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be prime"
        return int(self.pari_mincurve.ellap(p))

    def conductor(self):
        """
        Returns the conductor of the elliptic curve.
        """
        try:
            return self.__conductor
        except AttributeError:
            self.pari_mincurve # computes conductor as side effect
            return self.__conductor

    def kodaira_type(self, p):
        """
        Local Kodaira type of the elliptic curve at p.

        1 means good reduction (type I_0), 2, 3 and 4 mean types II, III and IV, respectively,
        4 + nu with nu > 0 means type I_{nu}; finally the opposite values -1, -2, etc. refer
        to the starred types I_0^*, II^*, etc.  
        """
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be prime"
        try:
            self.__kodaira_type
        except AttributeError:
            self.__kodaira_type = {}
            self.__tamagawa_number = {}
        if not self.__kodaira_type.has_key(p):
            v = self.pari_mincurve.elllocalred(p)
            self.__kodaira_type[p] = str(v[1])
            self.__tamagawa_number[p] = int(v[3])
        return self.__kodaira_type[p]

    def tamagawa_number(self, p):
        """
        The Tamagawa number of the elliptic curve at p.
        """
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be prime"
        try:
            return self.__tamagawa_number[p]
        except (AttributeError, KeyError):
            self.kodaira_type(p)
            return self.__tamagawa_number[p]
        
    def Lseries(self, s):
        """
        Returns the value of the L-series of the elliptic curve E at s, where s
        must be a real number.  

        NOTES:
        (1) If the conductor of the curve is large, say > 10^12, then this function
        will take a long time, since it uses an O(sqrt(N)) algorithm.
        
        (2) There is an algorithm to compute L(E,s) for any complex s, but it involves
        computing the incomplete Gamma function, which is not implemented in MANIN.

        EXAMPLES:
            >>> E=EllipticCurve([1,2,3,4,5])
            >>> E.Lseries(1)
            0.E-28

            The following may have low order bits that are different
            on your machine.
            >> E.Lseries(1.1)    
            0.28549100767814812844289234320
        """
        return self.pari_mincurve.elllseries(s)

    def sigma(self, z, flag=0):
        """
        Returns the value of the Weierstrass sigma function of the lattice
        associated to this elliptic curve E.
        INPUT:
            z -- a complex number
            flag -- 0 - default ???
                    1 - computes an arbitrary determination of log(sigma(z))
                    2, 3 - same using the product expansion instead of theta series.
                           ???
        OUTPUT:
            a complex number

        NOTE: The reason for the ???'s above, is that the PARI documentation for
              ellsigma is very vague.
        """
        return self.pari_curve.ellsigma(z, flag)

    def modular_parametrization(self):
        """
        Computes and returns ...
        """
        return self.pari_mincurve.elltaniyama()

    def cremona_label(self):
        try:
            return self.__cremona_label
        except AttributeError:
            raise errors.UsageError("Cremona label unknown.")

    def torsion_subgroup(self, flag=0):
        """
        Returns the torsion subgroup of this elliptic curve.
        """
        try:
            return self.__torsion
        except AttributeError:
            self.__torsion = self.pari_curve.elltors(flag)
        return self.__torsion

    def newform(self):
        """
        Returns the newform attached to this elliptic curve.
        """
        def comp(n):  # compute series expansion of newform attached E as power series
            v = self.anlist(n)
            return rings.PowerSeries(rings.PowerSeriesRing(rings.Q), v, n)
        M = modform.ModularForms(self.conductor(), 2)
        return modform.Newform(M, comp)


    def root_number(self):
        try:
            return self.__root_number
        except AttributeError:
            self.__root_number = int(self.pari_mincurve.ellrootno())
        return self.__root_number

# 	def	weierstrass_p(self):
#         # TODO: add allowing negative valuations for power series
#         return 1/t**2 + a1/t + rings.frac(1,12)*(a1-8*a2) -a3*t \
#                - (a4+a1*a3)*t**2  + O(t**3)

    def padic_sigma(self, p, prec=20):
        """
        Returns the p-adic sigma function of the elliptic curve
        as a power series in t to precision prec.
        """
        R = rings.PowerSeriesRing(variable="t")
        t = R.g0
        a1, a2, a3, a4, a6 = self.__ainvs
        misc.todo("Implement genuine computation of sigma.")
        return t + a1/2 * t**2 + (a1**2+a2)/3 * t**3 + (a1**3+2*a1*a2+2*a3)/4 * t**4 + O(t**5)

    def formal_w(self, prec=20):
        """
        Return the formal power series
        
                     w(t) = t^3 + ...

        to precision O(t^prec) of Proposition IV.1.1 of [Silverman
        AEC1].  This is the formal expansion of w = -1/y about the
        formal parameter t = -x/y at infinity.

        We compute w using the recursive procedure (4.1) on page 18
        of Bluher's `A leisurely introduction to formal groups and
        elliptic curves', which I downloaded from
        
            http://www.math.uiuc.edu/Algebraic-Number-Theory/0076/
            
        """
        try:
            pr, w = self.__formal_w
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = w.parent().gen()
            return w + O(t**prec)
        R = rings.PowerSeriesRing(self.base_field(), "t")
        t = R.gen()
        w = t**3 + O(t**prec)
        a1, a2, a3, a4, a6 = self.ainvs()
        for n in range(4,prec):
            if n<pr:
                w[n] = self.__formal_w[1][n]   # already know it
            else:
                w[n] = a1*w[n-1] + \
                       a2*w[n-2] + \
                       a3*sum([w[i]*w[n-i] for i in range(3,n-2)]) + \
                       a4*sum([w[i]*w[n-1-i] for i in range(3,n-3)]) + \
                       a6*sum([w[i]*w[j]*w[n-i-j] for i in range(3,n-2) \
                               for j in range(3,n-2) if n-i-j >= 3])
        self.__formal_w = (prec, w)
        return w

    def formal_x(self, prec=20):
        """
        Return the formal power series x(t) in terms of the local
        parameter t = -x/y at infinity.
        """
        prec = max(prec,0)
        try:
            pr, x = self.__formal_x
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = x.parent().gen()
            return x + O(t**prec)
        w = self.formal_w(prec+5)
        t = w.parent().gen()
        x = t/w   
        self.__formal_x = (prec, x)
        return x
    
    def formal_y(self, prec=20):
        """
        Return the formal power series y(t) in terms of the local
        parameter t = -x/y at infinity.
        """
        prec = max(prec,0)
        try:
            pr, y = self.__formal_y
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = y.parent().gen()
            return y + O(t**prec)
        w = self.formal_w(prec+6)
        t = w.parent().gen()
        y = -1/w   
        self.__formal_y = (prec, y)
        return y

    def formal_inverse(self, prec=20):
        prec = max(prec,0)
        try:
            pr, inv = self.__formal_inverse
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = inv.parent().gen()
            return inv + O(t**prec)
        x = self.formal_x(prec)
        y = self.formal_y(prec)
        a1, _, a3, _, _ = self.ainvs()
        inv = x / ( y + a1*x + a3)          # page 114 of Silverman, AEC I
        self.__formal_inverse = (prec, inv)
        return inv

    def formal_n_isogeny(self, n, prec=20):
        prec = max(prec,0)
        try:
            pr, phi = self.__formal_n_isogeny[n]
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = phi.parent().gen()
            return phi + O(t**prec)
        try:
            _ = self.__formal_n_isogeny
        except:
            self.__formal_n_isogeny = {}
        misc.todo()

    def formal_group(self, prec=10):
        prec = max(prec,0)
        try:
            pr, F = self.__formal_group
        except AttributeError:
            pr = -1
        if prec <= pr:
            return F
        R1 = rings.PowerSeriesRing(self.base_field(),"t1")
        R2 = rings.PowerSeriesRing(R1,"t2")
        t1 = R1.gen().add_bigoh(prec)
        t2 = R2.gen().add_bigoh(prec)
        w = self.formal_w(prec)
        def tsum(n):
            return sum([t2**m * t1**(n-m-1) for m in range(n)])
        lam = sum([tsum(n)*w[n] for n in range(3,prec)])
        print "lam = ", lam;
        w1 = R1(w.coeffs()).add_bigoh(prec)
        nu = w1 - lam*t1
        a1, a2, a3, a4, a6 = self.__ainvs
        lam2 = lam*lam
        lam3 = lam2*lam
        t3 = -t1 - t2 - \
             (a1*lam + a3*lam2 + a2*nu + 2*a4*lam*nu + 3*a6*lam2*nu)/  \
             (1 + a2*lam + a4*lam2 + a6*lam3)
        inv = self.formal_inverse(prec)
        F = inv(t3)
        self.__formal_group = (prec, F)
        return F

    def formal_mult(self, n, prec=10):
        R = rings.PowerSeriesRing(self.base_field(),"t")
        t = R._0
        if n == 1:
            return t +O(t**prec)
        if n == -1:
            return self.formal_inverse(prec)
        if n < 0:
            return self.formal_inverse(prec)(self.formal_mult(-n,prec))
        F = self.formal_group(prec)
        g = F.parent().base_ring().gen()
        for m in range(2,n+1):
            g = F(g)        
        return R(g.coeffs())

    def division_polynomial(self, n):
        misc.todo()

    def formal_log(self, prec=20):
        misc.todo()
                 

class Point:

    def __init__(self, curve, x=oo, y=oo):
        if not isinstance(curve, EllipticCurve):
            raise errors.TypeError("Argument 1 must be an elliptic curve.")
        self.__curve = curve
        if x is oo:
            self.__x, self.__y = oo, oo
            return
        K = curve.base_field()
        x = K(x)
        y = K(y)
        self.__x, self.__y = x, y
        if y*y + curve.a1()*x*y + curve.a3()*y != \
               x**3 + curve.a2()*x**2 + curve.a4()*x + curve.a6():
            raise errors.MathError("Point %s is not on curve."%self)

    def __repr__(self):
        if self.is_infinity():
            return "0"
        return "(%s, %s)"%(self.__x,self.__y)

    def order(self):
        """
        Return the order of this point on the elliptic curve.
        If the point has infinite order, returns 0.

        EXAMPLE:
            >>> E = EllipticCurve([0,0,1,-1,0])
            >>> P = E([0,0]); P
            (0, 0)
            >>> P.order()
            0

            >>> E = EllipticCurve([0,1])
            >>> P=E([-1,0])
            >>> P.order()
            2
        """
        if self.is_infinity():
            return 1
        return self.__curve.pari_curve.ellorder([self.__x, self.__y])

    def xy(self):
        return self.__x, self.__y

    def x(self):
        return self.__x

    def y(self):
        return self.__y

    def curve(self):
        return self.__curve

    def __getitem__(self, n):
        if n==0:
            return self.__x
        elif n==1:
            return self.__y
        else:
            raise IndexError, "Index must be 0 or 1"

    def is_infinity(self):
        return self.__x is rings.infinity
    
    def __add__(self, right):
        """Add self to right."""
        if self.curve() != right.curve():
            raise errors.MathError("Points must be on the same curve.")

        # Use Prop 7.1.7 of Cohen "A Course in Computational Algebraic Number Theory"
        if self.is_infinity():
            return right
        if right.is_infinity():
            return self
        a1, a2, a3, a4, a6 = self.__curve.ainvs()
        x1, y1 = self.__x, self.__y
        x2, y2 = right.__x, right.__y
        if x1 == x2 and y1 == -y2 - a1*x2 - a3:
            return Point(self.__curve) # point at infinity
        if x1 == x2 and y1 == y2:
            m = (3*x1*x1 + 2*a2*x1 + a4 - a1*y1) / (2*y1 + a1*x1 + a3)
        else:
            m = (y1-y2)/(x1-x2)
        x3 = -x1 - x2 - a2 + m*(m+a1)
        y3 = -y1 - a3 - a1*x3 + m*(x1-x3)
        return Point(self.__curve, x3, y3)

    def __sub__(self, right):
        return self + (-right)

    def __neg__(self):
        if self.is_infinity():
            return self
        E, x, y = self.__curve, self.__x, self.__y
        return Point(self.__curve, x, -y - E.a1()*x - E.a3())

    def __rmul__(self, m):
        if m<0:
            return (-m)*(-self)
        if m==1:
            return self
        if m==0:
            return Point(self.__curve)
        power = Point(self.__curve)
        i = 0
        apow2 = self
        while ((m>>i) > 0):
            if((m>>i) & 1):
                power += apow2
            apow2 += apow2
            i += 1
        return power

    def sigma(self, p):
        k = rings.Qp(p)
        if self.is_infinity():
            return k(0)
        sigma = self.__curve.sigma(p)
        return sigma(k(-self.__x/self.__y))
        

    def padic_height(self, p):
        """Returns the p-adic cyclotomic height of the point.

        padic_height(p)

        Input:
           p: a prime number
        Output:
           p-adic cyclotomic height, which is a single p-adic number.

        Algorithm:
        We compute this height using the following formula, which
        is valid for points that are in the intersection of the
        identity component of the Neron model with the kernel of
        reduction modulo p:

     h(P) = 1/2 * sum_{ell!=p} sup(0,-ord_ell(x(P)))  + log_p(sigma_p(-x(P)/y(P)) / e),

        where P=(a/e^2, b/e^3) with gcd(a,e)=1, and
        where the first sum is over primes ell that don't equal p.
        If P isn't in the subgroup mentioned above, let n be a positive
        integer so that n*P is in that subgroup.  Then we return
        h(n*P)/(n**2), which does not depend on the choice of n, and
        is defined using the above formula.

        """

        if not arith.is_prime(p):
            raise errors.UsageError("p must be a prime number.")
        E = self.curve()
        if E.conductor() % p == 0:
            raise errors.UsageError("p must be a prime of good reduction.")

        ap = E.ap(p)
        if ap == 0:
            raise errors.UsageError("p must a prime of good ordinary reduction.")
        
        Np = p+1 - E.ap(p)
        c = 1
        for q, _ in arith.factor(E.conductor()):
            c = arith.lcm(c, E.tamagawa_number(q))

        n = Np*c
        misc.verbose("Multiplying point by %s."%n)
        R = n*self
            
        x, y = R.xy()
        sigma = E.padic_sigma(p)
        Qp = rings.pAdicField(p)
        print "x = ", x
        print "-x/y = ", Qp(-x/y)
        d = x.denominator()
        e = d.sqrt()
        if e*e != d:
            raise errors.MathError("The denominator is not a perfect square!?")
        t = Qp(-x/y)
        return (sigma(t)/Qp(e)).log()



        w = Qp(0)
        for ell, ord in d.factor():
            if ell != p:
                w += ord * Qp(ell).log()
        # note that there is no 1/2, since we factored the square root.
        t = Qp(-x/y)
        print "sigm = ", sigma(t)
        print "e = ", e
        hP = w + (sigma(t)/e).log()
        print "w = ", w
        print "sigma(t) = ", sigma(t)
        print "sigma(t)/e = ", sigma(t)/e
        print "(sigma(t)/e).log() = ", (sigma(t)/e).log()
        print "sum = ", w + (sigma(t)/e).log()

        return hP/Qp(n**2)



                
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
