"""
Miscellaneous functions for timing, verbosity, etc.
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator, os, sys, time, weakref, random

SAGE_HOME = os.environ["SAGE_HOME"]
LOGFILE = "%s/log/sage_log"%SAGE_HOME


#################################################################
# Functions to help with interfacing with CXX code that
# uses the GMP library
#################################################################
def to_gmp_hex(n):
    return hex(n).replace("L","").replace("0x","")

def cputime(t=0):
    if t==0:
        return time.clock()
    return time.clock() - t

def clock(cmd):
    t=cputime()
    eval(compile(cmd,"clock",'single'))
    return cputime(t)

LEVEL=0  # default
def verbose(mesg="", t=0, level=1, caller_name=None):
    if level>LEVEL:
        return cputime()
    if t != 0 and mesg=="":
        mesg = "Finished."
    # see recipe 14.7 in Python Cookbook
    if caller_name == None:
        caller_name = sys._getframe(1).f_code.co_name + ": "
        if caller_name == "?: ":
            caller_name = ""
    s = "%s%s"%(caller_name, mesg)
    if t!=0:
        s = s + " (time = %s)"%cputime(t)
    #print s
    open(LOGFILE,"a").write(s+"\n")
    return cputime()

def todo(mesg=""):
    caller_name = sys._getframe(1).f_code.co_name
    print "%s: todo -- %s"%(caller_name, mesg)

def set_verbose(level):
    """
    Set the verbosity level.

    INPUT:
       int level: an integer between 0 and 2, inclusive.
    OUTPUT:
       changes the state of the verbosity flag.
    """
    global LEVEL
    LEVEL = level

def get_verbose():
    global LEVEL
    return LEVEL

def generic_cmp(x,y):
    if x<y:
        return -1
    elif x==y:
        return 0
    return 1
    
def mul(x, z=None):
    """
    Return the product of the elements in the list x.
    If optimal argument z is not given, start the product
    with the first element of the list, otherwise use z.

    >>> mul([1,2,34])
    68
    >>> mul([2,3], 5)
    30
    """
    if z == None:
        if len(x) == 0:
            return 1
        elif len(x) == 1:
            return x[0]
        return reduce(operator.mul, x[1:], x[0])
    else:
        if len(x) == 0:
            return z
        return reduce(operator.mul, x, z)

# alternative name for mul
prod = mul

def add(x, z=0):
    return reduce(operator.add, x, z)

def add_nonempty(x):
    return reduce(operator.add, x[1:], x[0])

def union(x, y=None):
    if y==None:
        if len(x) == 0:
            return []
        z = x[0]
        for i in range(1,len(x)):
            z = union(z,x[i])
        return z
    else:
        z = x + y
        z.sort()
        w=[]
        for i in range(len(z)):
            if i==len(z)-1 or z[i]!=z[i+1]:
                w.append(z[i])
        return w

def uniq(x):
    """
    Return a sublist of the elements in the list x that is sorted and
    is such that the entries in the sublist are unique.
    """
    assert isinstance(x,list), "Argument must be a list."
    d={}
    for a in x:
        d[a] = None
    return d.keys()
        
# Compute a nice string representation of a linear
# combination of some formal symbols
def repr_lincomb(symbols, lincomb):
    s = ""
    first = True
    i=0
    for c in lincomb:
        b = symbols[i]
        if c != 0:
            if c == 1:
                if not first:
                    coeff = " + "
                else:
                    coeff = ""
            elif c == -1:
                if not first:
                    coeff = " - "
                else:
                    coeff = "-"
            else:
                if not first:
                    coeff = " + %s*"%c
                else:
                    coeff = "%s*"%c
            s += "%s%s"%(coeff, b)
            first = False
        i += 1
    if first:
        s = "0"
#     else:
#         if isinstance(lincomb[0], (int, long, rings.Integer, rings.Rational)):
#             s = s.replace("+ -","- ")
    return s

    
#################################################################
# Generally useful
#################################################################

def parent(x):
    """Returns the parent of x, if x has a parent, otherwise
    return None."""
    try:
        return x.parent()
    except AttributeError:
        return None


def assert_attribute(x, attr, init=None):
    """
    If the object x has the attribute attr, do nothing.
    If not, set x.attr to init.
    """
    if x.__dict__.has_key(attr): return
    if attr[:2] == "__":
        z = str(x.__class__).split("'")
        if len(z) > 1:
            z = z[1]
        else:
            z = z[0]
        attr = "_" + z[len(x.__module__)+1:] + attr
    x.__dict__[attr] = init

#################################################################
# Useful but hard to classify
#################################################################

def random_sublist(X, s):
    """
    Return a random sublist of the list X where the chance
    of including a particular element is s.

    INPUT:
        X -- list
        s -- floating point number between 0 and 1
    OUTPUT:
        list
    """
    return [a for a in X if random.random() <= s]
