"""
Univariate Polynomial Rings
"""

import random
import weakref

from ring import Ring
import sage.structure.gens as gens
import polynomial 
import rational_field

_objsPolynomialRing = {}
class _uniqPolynomialRing(object):
    def __new__(cls, base_ring, variable="x", sparse=False):
        key = (base_ring, variable, sparse)
        if _objsPolynomialRing.has_key(key):
            x = _objsPolynomialRing[key]()
            if x != None: return x
        R = object.__new__(cls)
        _objsPolynomialRing[key] = weakref.ref(R)
        return R

class PolynomialRing(_uniqPolynomialRing, Ring, gens.Gens):
    """
    Univariate polynomial ring.
    
    EXAMPLES:
        >>> import rings
        >>> Q = rings.RationalField()
        >>> R = PolynomialRing(Q)
        >>> x = R.gen()
        >>> R(-1) + R(1)
        0
        >>> (x - '2/3')*(x**2 - 8*x + 16)
        x^3 - 26/3*x^2 + 64/3*x - 32/3
    """
    
    def __init__(self, base_ring, variable="x", sparse=False):
        self.__base_ring = base_ring
        self.__variable = variable
        self.__is_sparse = sparse
        Ring.__init__(self)
        self.__set_polynomial_class()
        self.__generator = self.polynomial([0,1], is_gen=True)
        self.__cyclopoly_cache = {}
        
    def __call__(self, *args, **kwds):
        return self.polynomial(*args, **kwds)

    def __cmp__(self, other):
        if not isinstance(other, PolynomialRing):
            return -1
        if self.__variable == other.__variable:
            return 0
        elif self.__variable < other.__variable:
            return -1
        return +1
        
    def __contains__(self, x):
        return isinstance(x, polynomial.Polynomial) and x.parent() == self
        
    def __repr__(self):
        s = "Univariate Polynomial Ring in %s over %s"%(
                self.__variable, self.__base_ring)
        if self.is_sparse():
            s = "Sparse " + s
        return s
        
    def __set_polynomial_class(self):
        if self.__is_sparse:
            self.__polynomial_class = polynomial.Polynomial_generic_sparse
        else:
            if isinstance(self.base_ring(), rational_field.RationalField):
                self.__polynomial_class = polynomial.Polynomial_rational_dense
            else:
                self.__polynomial_class = polynomial.Polynomial_generic_dense
            
    def base_ring(self):
        return self.__base_ring
            
    def characteristic(self):
        return self.__base_ring.characteristic()
        
    def cyclotomic_polynomial(self, n):
        """
        The nth cyclotomic polynomial, which is irreducible and
        has a primitive nth root of unity as root.
        """
        if n<=0:
            raise errors.UsageError("n=%s must be positive")
        if self.__cyclopoly_cache.has_key(n):
            return self.__cyclopoly_cache[n]
        
        import sage.pari.pari_pyx as pari
        x = self.gen(0)
        self.__cyclopoly_cache[n] = self(eval(pari.eval(\
            "f=polcyclo(%s); vector(poldegree(f)+1,i,polcoeff(f,i-1))"%n)))
        return self.__cyclopoly_cache[n]
        
        # todo: this desperately needs to be optimized (using PARI above is "lame")
        if n == 1:
            self.__cyclopoly_cache[n] = self.gen()-1
            return self.__cyclopoly_cache[1]
        if arith.is_prime(n):
            R = self.base_ring()
            f = self([R(1) for _ in range(n)])
            self.__cyclopoly_cache[n] = f
            return f
        F = arith.factor(n)
        if len(F) == 1:
            p = F[0][0]
            r = F[0][1]
            R = self.base_ring()
            pow = p**(r-1)
            v = [R(0) for _ in range((p-1)*pow+1)]
            for i in range(p):
                v[i*pow] = R(1)
            h = self(v)
            self.__cyclopoly_cache[n] = h
            return h
        h = self(1)
        for d in arith.divisors(n):
            if d != n:
                h *= self.cyclotomic_polynomial(d)
        f = self.gen()**n - 1
        q,r = f/h
        assert r.is_zero()
        self.__cyclopoly_cache[n] = q
        return q
        

    def gen(self, n=0):
        """
        If this is R[x], this intrinsic returns x.
        """
        if n != 0:
            raise IndexError, "Generator %s not defined."%n
        return self.__generator
            
    def is_field(self):
        return False
        
    def is_sparse(self):
        return self.__is_sparse
        
    def name(self):
        return "poly-%s-%s"%(self.__variable,self.__base_ring)
    
    def ngens(self):
        return 1
        
    def polynomial(self, *args, **kwds):
        return self.__polynomial_class(self, *args, **kwds)
    
    def random(self, degree, bound=10):
        """
        Return a random polynomial.
        
        INPUT:
            degree -- an int
            bound -- an int
        OUTPUT: Polynomial
           A polynomial whose coefficients of x^i, for i up to degree,
           are coercisions to the base ring of random integers between
           -bound and bound.
        """
        return self([random.randrange(-bound,bound) for _ in xrange(degree+1)])

    def variable(self):
        """
        Returns the string which is used to print the generator of the polynomial ring.
        OUTPUT:
            str -- generator name
        """
        return self.__variable

    def set_variable(self, variable):
        self.__variable = variable
    

# doctests
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
    
