"""
Pyrex support for rings
"""


#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import manin.rings.rational_field as rational_field
import manin.rings.integer_ring as integer_ring
import manin.misc.misc as misc
import manin.rings.arith_c
inverse_mod_int = manin.rings.arith_c.inverse_mod_int

include "cdefs.pxi"
include "gmp.pxi"

cdef class Integer
cdef class Rational

cdef set_mpz(Integer self, mpz_t value):
    mpz_set(self.value, value)

cdef class Integer:
    cdef mpz_t value

    def __new__(self, x=None):
        mpz_init(self.value)

    def __init__(self, x=None):
        if x != None:
            s = str(x)
            mpz_set_str(self.value, s, 0)

    def __cmp__(Integer self, Integer x):
        cdef int i
        i = mpz_cmp(self.value, x.value)
        if i < 0:
            return -1
        elif i == 0:
            return 0
        else:
            return 1

    def copy(self):
        cdef Integer z
        z = Integer()
        set_mpz(z,self.value)
        return z

    def  __dealloc__(self):
        mpz_clear(self.value)
        
    def __repr__(self):
        cdef char *s
        s = mpz_get_str(NULL, 10, self.value)
        t = str(s)
        free(s)
        return t
    
    def set_si(self, signed long int n):
        mpz_set_si(self.value, n)

    def set_str(self, s, base=10):
        valid = mpz_set_str(self.value, s, base)
        if valid != 0:
            raise ValueError, "invalid literal:" + s

    def __add__(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_add(x.value, self.value, other.value)
        return x

    def __sub__(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_sub(x.value, self.value, other.value)
        return x

    def __mul__(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_mul(x.value, self.value, other.value)
        return x
    
    def __div__(Integer self, Integer other):
        cdef Rational y
        cdef Integer x
        if mpz_divisible_p(self.value, other.value):
            x = Integer()
            mpz_divexact(x.value, self.value, other.value)
            return x
        else:
            y = Rational()
            mpq_set_num(y.value, self.value)
            mpq_set_den(y.value, other.value)
            mpq_canonicalize(y.value)
            return y


    def __pow__(Integer self, long int n, dummy):
        cdef Integer x
        x = Integer()
        mpz_pow_ui(x.value, self.value, n)
        return x

    def __pos__(self):
        return self

    def __neg__(self):
        cdef Integer x
        x = Integer()
        mpz_neg(x.value, self.value)
        return x

    def __abs__(self):
        cdef Integer x
        x = Integer()
        mpz_abs(x.value, self.value)
        return x

    def __mod__(Integer self, Integer modulus):
        cdef Integer x
        x = Integer()
        mpz_mod(x.value, self.value, modulus.value)
        return x

    def powermod(self, Integer exp, Integer mod):
        """
        powermod(self, Integer exp, Integer mod):
        Compute self**exp modulo mod.
        """
        cdef Integer x
        x = Integer()
        mpz_powm(x.value, self.value, exp.value, mod.value)
        return x

    def powermodm_ui(self, unsigned long int exp, Integer mod):
        cdef Integer x
        x = Integer()
        mpz_powm_ui(x.value, self.value, exp, mod.value)
        return x

    def __int__(self):
        return int(long(self))

    def __long__(self):
        cdef char *s
        s = mpz_get_str(NULL, 16, self.value)
        t = "0x%s"%str(s)
        free(s)
        return eval(t)

    def factor(self):
        return integer_ring.factor(self)

    def lcm(self, Integer n):
        cdef mpz_t x
        mpz_init(x)
        mpz_lcm(x, self.value, n.value)
        cdef Integer z
        z = Integer()
        mpz_set(z.value,x)
        mpz_clear(x)
        return z

    def is_one(self):
        return mpz_cmp_si(self.value, 1) == 0

    def is_zero(self):
        return mpz_cmp_si(self.value, 0) == 0

    def parent(self):
        return integer_ring.Z

    def xgcd(self, Integer n):
        """
        Return a triple g, s, t of Integers such that
               g = s * self + t * n.
        """
        cdef mpz_t g, s, t
        cdef object g0, s0, t0
        mpz_init(g)
        mpz_init(s)
        mpz_init(t)
        mpz_gcdext(g, s, t, self.value, n.value)
        g0 = Integer()
        s0 = Integer()
        t0 = Integer()
        set_mpz(g0,g)
        set_mpz(s0,s)
        set_mpz(t0,t)
        mpz_clear(g)
        mpz_clear(s)
        mpz_clear(t)
        return g0, s0, t0

    def gcd(self, Integer n):
        """
        Return the greatest common divisor of self and n.
        """
        cdef mpz_t g
        cdef object g0
        mpz_init(g)
        mpz_gcd(g, self.value, n.value)
        g0 = Integer()
        set_mpz(g0,g)
        mpz_clear(g)
        return g0

    def crt(self, Integer y, Integer m, Integer n):
        """
        Return unique integer between 0 and m*n that is self modulo m
        and y modulo n.  We assume that m and n are coprime.
        """
        cdef object g, s, t
        g, s, t = m.xgcd(n)
        if not g.is_one():
            raise ArithmeticError, "CRT requires that gcd of moduli is 1."
        # Now s*m + t*n = 1, so the answer is x + (y-x)*s*m, where x=self.
        return (self + (y-self)*s*m) % (m*n)

    def rational_reconstruction(self, Integer m):
        """
        View self as an integer modulo n and use rational reconstruction
        to try to lift to Q.
        INPUT:
            self -- Integer
            m -- Integer
        OUTPUT:
            x -- Rational
        """
        cdef mpq_t z
        cdef object x
        mpq_init(z)
        mpq_rational_reconstruction(z, self.value, m.value)
        x = Rational()
        set_mpq(x, z)
        mpq_clear(z)
        return x

def factorial(unsigned long int n):
    cdef mpz_t x
    cdef Integer z
    
    mpz_init(x)
    mpz_fac_ui(x, n)
    z = Integer()
    set_mpz(z, x)
    mpz_clear(x)
    return z

cdef object Rational_mul(Rational a, Rational b):
    cdef Rational x
    x = Rational()
    mpq_mul(x.value, a.value, b.value)
    return x

cdef object Rational_div(Rational a, Rational b):
    cdef Rational x
    x = Rational()
    mpq_div(x.value, a.value, b.value)
    return x

cdef Rational_add(Rational self, Rational other):
    cdef Rational x
    x = Rational()
    mpq_add(x.value, self.value, other.value)
    return x

cdef Rational_sub(Rational self, Rational other):
    cdef Rational x
    x = Rational()
    mpq_sub(x.value, self.value, other.value)
    return x

cdef set_mpq(Rational self, mpq_t value):
    mpq_set(self.value, value)

cdef class Rational:
    cdef mpq_t value

    def __new__(self, x=None):
        mpq_init(self.value)        

    def __init__(self, x=None):
        if x != None:
            s = str(x)
            mpq_set_str(self.value, s, 0)
            mpq_canonicalize(self.value)

    def __cmp__(Rational self, Rational x):
        cdef int i
        i = mpq_cmp(self.value, x.value)
        if i < 0:
            return -1
        elif i == 0:
            return 0
        else:
            return 1

    def copy(self):
        cdef Rational z
        z = Rational()
        mpq_set(z.value, self.value)
        return z

    def  __dealloc__(self):
        mpq_clear(self.value)
        
    def __repr__(self):
        cdef char *s
        s = mpq_get_str(NULL, 10, self.value)
        t = str(s)
        free(s)
        return t
    
    def set_si(self, signed long int n):
        mpq_set_si(self.value, n, 1)

    def set_str(self, s, base=10):
        valid = mpq_set_str(self.value, s, base)
        if valid != 0:
            raise ValueError, "invalid literal:" + s

    cdef set_mpz(self, mpz_t value):
        mpq_set_z(self.value, value)

    def __add__(Rational self, Rational other):
        cdef Rational x
        x = Rational()
        mpq_add(x.value, self.value, other.value)
        return x

    def __sub__(Rational self, Rational other):
        cdef Rational x
        x = Rational()
        mpq_sub(x.value, self.value, other.value)
        return x

    def __mul__(Rational self, Rational other):
        cdef Rational x
        x = Rational()
        mpq_mul(x.value, self.value, other.value)
        return x
    
    def __div__(Rational self, Rational other):
        cdef Rational x
        x = Rational()
        mpq_div(x.value, self.value, other.value)
        return x

    def __pow__(Rational self, long int n, dummy):
        cdef Rational x
        x = Rational()
        cdef mpz_t num, den
        mpz_init(num)
        mpz_init(den)
        mpz_pow_ui(num, mpq_numref(self.value), n)
        mpz_pow_ui(den, mpq_denref(self.value), n)
        mpq_set_num(x.value, num)
        mpq_set_den(x.value, den)
        mpz_clear(num)
        mpz_clear(den)        
        return x

    def __pos__(self):
        return self

    def __neg__(self):
        cdef Rational x
        x = Rational()
        mpq_neg(x.value, self.value)
        return x

    def __abs__(self):
        cdef Rational x
        x = Rational()
        mpq_abs(x.value, self.value)
        return x

    def mod_ui(Rational self, unsigned long int n):
        cdef int num, den, a
        cdef Rational x
        num = mpz_fdiv_ui(mpq_numref(self.value), n)
        den = mpz_fdiv_ui(mpq_denref(self.value), n)
        a = inverse_mod_int(den, n)
        return (a*num)%n

    def parent(self):
        return rational_field.Q

    def numer(self):
        cdef Integer n
        n = Integer()
        mpz_set(n.value, mpq_numref(self.value))
        return n
        
    def denom(self):
        cdef Integer d
        d = Integer()
        mpz_set(d.value, mpq_denref(self.value))
        return d
        
    def height(self):
        """
        The max absolute value of the numerator and denominator of self,
        as an Integer.
        """
        x = abs(self.numer())
        if x > self.denom():
            return x
        return self.denom()
        
    def order(self):
        if mpz_cmp_si(mpq_denref(self.value), 1) != 0:
            return 0
        if mpz_cmp_si(mpq_numref(self.value), 1) == 0:
            return 1
        elif mpz_cmp_si(mpq_numref(self.value), -1) == 0:
            return 2
        return 0
