"""
The integers.
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

import integer_ring 
import arith_c
import coerce

include "cdefs.pxi"
include "gmp.pxi"

cimport element
import element

cdef class Integer(element.Element)

import rational

Rational = rational.Rational

cdef public int set_mpz(Integer self, mpz_t value):
    mpz_set(self.value, value)

cdef set_from_Integer(Integer self, Integer other):
    mpz_set(self.value, other.value)

cdef set_from_int(Integer self, int other):
    mpz_set_si(self.value, other)

cdef public mpz_t* get_value(Integer self):
    return &self.value

cdef class Integer(element.Element):
    cdef mpz_t value

    def __new__(self, x=None):
        mpz_init(self.value)

    def __init__(self, x=None):
        if x != None:
            if isinstance(x, Integer):
                set_from_Integer(self, x) 
            elif isinstance(x, int):
                mpz_set_si(self.value, x)
            elif isinstance(x, long):
                s = "%x"%x
                mpz_set_str(self.value, s, 16)                
            elif isinstance(x, str):
                if x.isdigit() or x[1:].isdigit():
                    mpz_set_str(self.value, x, 0)
            elif isinstance(x, Rational):
                set_from_Integer(self, x.numer())
            else:
                raise TypeError, "Unable to coerce %s to an Integer."%x     

    cdef int cmp(self, Integer x):
        cdef int i
        i = mpz_cmp(self.value, x.value)
        if i < 0:
            return -1
        elif i == 0:
            return 0
        else:
            return 1

    def __cmp__(Integer self, Integer x):
        return self.cmp(x)

    def __richcmp__(Integer self, x, int op):
        cdef int n
        if not isinstance(x, Integer):
            x = Integer(x)
        n = self.cmp(x)
        if op == 0:
            return bool(n < 0)
        elif op == 1:
            return bool(n <= 0)
        elif op == 2:
            return bool(n == 0)
        elif op == 3:
            return bool(n != 0)
        elif op == 4:
            return bool(n > 0)
        elif op == 5:
            return bool(n >= 0)

    def copy(self):
        """
        Return a copy of the integer.
        """
        cdef Integer z
        z = Integer()
        set_mpz(z,self.value)
        return z

    def  __dealloc__(self):
        mpz_clear(self.value)
        
    def __repr__(self):
        return self.str()
        
    def str(self, int base=10):
        cdef char *s
        s = mpz_get_str(NULL, base, self.value)
        t = str(s)
        free(s)
        return t
    
    def set_si(self, signed long int n):
        """
        Coerces n to a C signed integer if possible, and sets self
        equal to n.
        """
        mpz_set_si(self.value, n)

    def set_str(self, s, base=10):
        """
        Set self equal to the number defined by the s in the given
        base.
        """
        valid = mpz_set_str(self.value, s, base)
        if valid != 0:
            raise ValueError, "invalid literal:" + s

    def add(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_add(x.value, self.value, other.value)
        return x
        
    def __add__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.add(y)
        return coerce.bin_op(x, y, operator.add)

    def sub(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_sub(x.value, self.value, other.value)
        return x
        
    def __sub__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.sub(y)
        return coerce.bin_op(x, y, operator.sub)

    def mul(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_mul(x.value, self.value, other.value)
        return x
            
    def __mul__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.mul(y)
        return coerce.bin_op(x, y, operator.mul)
        
    def div(Integer self, Integer other):
        cdef Integer x
        if mpz_divisible_p(self.value, other.value):
            x = Integer()
            mpz_divexact(x.value, self.value, other.value)
            return x
        else:
            return rational.Rational(self)/rational.Rational(other)
        #raise ArithmeticError, "Exact division impossible."
        
    def __div__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.div(y)
        return coerce.bin_op(x, y, operator.div)

    def __pow__(Integer self, long int n, dummy):
        cdef Integer x
        x = Integer()
        mpz_pow_ui(x.value, self.value, n)
        return x

    def __pos__(self):
        return self

    def __neg__(self):
        cdef Integer x
        x = Integer()
        mpz_neg(x.value, self.value)
        return x

    def __abs__(self):
        cdef Integer x
        x = Integer()
        mpz_abs(x.value, self.value)
        return x

    def __mod__(Integer self, Integer modulus):
        cdef Integer x
        x = Integer()
        mpz_mod(x.value, self.value, modulus.value)
        return x

    def powermod(self, Integer exp, Integer mod):
        """
        powermod(self, Integer exp, Integer mod):
        Compute self**exp modulo mod.
        """
        cdef Integer x
        x = Integer()
        mpz_powm(x.value, self.value, exp.value, mod.value)
        return x

    def powermodm_ui(self, unsigned long int exp, Integer mod):
        cdef Integer x
        x = Integer()
        mpz_powm_ui(x.value, self.value, exp, mod.value)
        return x

    def __int__(self):
        cdef char *s
        s = mpz_get_str(NULL, 16, self.value)
        n = int(s,16)
        free(s)
        return n

    def __long__(self):
        cdef char *s
        s = mpz_get_str(NULL, 16, self.value)
        n = long(s,16)
        free(s)
        return n

    def __float__(self):
        return mpz_get_d(self.value)

    def factor(self):
        """
        Return the prime factorization of self as a list of
        pairs (p,e).
        """
        return integer_ring.factor(self)

    def lcm(self, Integer n):
        """
        Returns the least common multiple of self and n.
        """
        cdef mpz_t x
        mpz_init(x)
        mpz_lcm(x, self.value, n.value)
        cdef Integer z
        z = Integer()
        mpz_set(z.value,x)
        mpz_clear(x)
        return z
        
    def denominator(self):
        return ONE

    def is_one(self):
        return mpz_cmp_si(self.value, 1) == 0

    def is_zero(self):
        return mpz_cmp_si(self.value, 0) == 0

    def parent(self):
        """
        Return IntegerRing().
        """
        return integer_ring.Z

    def sqrt(self):
        """
        Returns the integer floor of the square root of self.
        EXAMPLE:
            >>> a = Integer(5)
            >>> a.sqrt()
            2
        """
        cdef Integer x
        x = Integer()
        mpz_sqrt(x.value, self.value)
        return x

    def xgcd(self, Integer n):
        """
        Return a triple g, s, t of Integers such that
               g = s * self + t * n.
        """
        cdef mpz_t g, s, t
        cdef object g0, s0, t0
        
        mpz_init(g)
        mpz_init(s)
        mpz_init(t)
        mpz_gcdext(g, s, t, self.value, n.value)
        g0 = Integer()
        s0 = Integer()
        t0 = Integer()
        set_mpz(g0,g)
        set_mpz(s0,s)
        set_mpz(t0,t)
        mpz_clear(g)
        mpz_clear(s)
        mpz_clear(t)
        return g0, s0, t0

    def inverse_mod(self, n):
        """
        Returns the inverse of self modulo n, if this inverse exists.
        Otherwise, raises a ZeroDivisionError exception.
        
        INPUT:
           self -- Integer
           m -- Integer
        OUTPUT:
           x -- Integer such that x*self = 1 (mod m), or
                raises ZeroDivisionError.
        IMPLEMENTATION:
           Call the mpz_invert GMP library function.
        EXAMPLES:
        >>> a = Integer(189)
        >>> a.inverse_mod(10000)
        4709
        >>> a.inverse_mod(-10000)
        4709
        >>> a.inverse_mod(1890)
        Traceback (most recent call last):
        ...
        ZeroDivisionError: Inverse of 189 mod 1890 does not exist.
        """
        cdef mpz_t x
        cdef object ans
        cdef int r
        cdef Integer m
        m = Integer(n)
        
        mpz_init(x)
        r = mpz_invert(x, self.value, m.value)
        if r == 0:
            raise ZeroDivisionError, "Inverse of %s mod %s does not exist."%(self,m)
        ans = Integer()
        set_mpz(ans,x)
        mpz_clear(x)
        return ans
        
    def gcd(self, Integer n):
        """
        Return the greatest common divisor of self and n.
        """
        cdef mpz_t g
        cdef object g0
        mpz_init(g)
        mpz_gcd(g, self.value, n.value)
        g0 = Integer()
        set_mpz(g0,g)
        mpz_clear(g)
        return g0

    def crt(self, Integer y, Integer m, Integer n):
        """
        Return unique integer between 0 and m*n that is self modulo m
        and y modulo n.  We assume that m and n are coprime.
        """
        cdef object g, s, t
        g, s, t = m.xgcd(n)
        if not g.is_one():
            raise ArithmeticError, "CRT requires that gcd of moduli is 1."
        # Now s*m + t*n = 1, so the answer is x + (y-x)*s*m, where x=self.
        return (self + (y-self)*s*m) % (m*n)
        
ONE = Integer(1)

def factorial(unsigned long int n):
    """
    Return the factorial 1 *2 * 3 * n.
    """
    cdef mpz_t x
    cdef Integer z
    
    mpz_init(x)
    mpz_fac_ui(x, n)
    z = Integer()
    set_mpz(z, x)
    mpz_clear(x)
    return z

def _test():
    import doctest, sys
    doctest.testmod(sys.modules[__name__], verbose=True)

if __name__ ==  '__main__':
    _test()
