"""
The class of rational numbers.
"""


#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

import rational_field
import manin.misc.misc as misc
import arith_c
import coerce

##cimport integer
import integer

cimport element
import element

inverse_mod_int = arith_c.inverse_mod_int

include "cdefs.pxi"
include "gmp.pxi"


# This is tight coupling, in that integer must be a c-type and be
# compiled before rational.  Change numer and denom below to
# something slightly less efficient to eliminate this.
cdef extern from "integer.h":
    int set_mpz(void *, mpz_t value)
    mpz_t* get_value(void *)

cdef class Rational(element.Element)

cdef public void set_from_mpq(Rational self, mpq_t value):
    mpq_set(self.value, value)
    
cdef public void set_from_Rational(Rational self, Rational other):
    mpq_set(self.value, other.value)

##cdef public void set_from_Integer(Rational self, integer.Integer other):
##    mpz_set(mpq_numref(self.value), other.value)
##    mpz_set_si(mpq_denref(self.value), 1)

cdef public void set_from_Integer(Rational self, other):
    s = other.str(16)
    mpq_set_str(self.value, s, 16)

cdef object Rational_mul(Rational a, Rational b):
    cdef Rational x
    x = Rational()
    mpq_mul(x.value, a.value, b.value)
    return x

cdef object Rational_div(Rational a, Rational b):
    cdef Rational x
    x = Rational()
    mpq_div(x.value, a.value, b.value)
    return x

cdef Rational_add(Rational self, Rational other):
    cdef Rational x
    x = Rational()
    mpq_add(x.value, self.value, other.value)
    return x

cdef Rational_sub(Rational self, Rational other):
    cdef Rational x
    x = Rational()
    mpq_sub(x.value, self.value, other.value)
    return x
    
cdef class Rational(element.Element):
    cdef mpq_t value

    def __new__(self, x=None):
        mpq_init(self.value)

    def __init__(self, x=None):
        cdef int i
        if x != None:
            self.__set_value(x)

    def __set_value(self, x):
        if isinstance(x, Rational):
            set_from_Rational(self, x)
        elif isinstance(x, int):
            i = x
            mpq_set_si(self.value, i, 1)
        elif isinstance(x, long):
            s = "%x"%x
            mpq_set_str(self.value, s, 16)
        elif isinstance(x, integer.Integer):
            set_from_Integer(self, x)
        elif isinstance(x, str):
            mpq_set_str(self.value, x, 0)
            mpq_canonicalize(self.value)
        elif isinstance(x, tuple) and len(x) == 2:
            n, d = x
            s = "%s/%s"%(n,d)
            mpq_set_str(self.value, s, 0)
            mpq_canonicalize(self.value)
        elif isinstance(x, list) and len(x) == 1:
            self.__set_value(x[0])
        else:
            raise TypeError, "can't coerce %s to Rational"%str(x)

    cdef cmp(Rational self, Rational x):
        cdef int i
        i = mpq_cmp(self.value, x.value)
        if i < 0:
            return -1
        elif i == 0:
            return 0
        else:
            return 1

    def __cmp__(self, x):
        return self.cmp(x)

    def __richcmp__(Rational self, x, int op):
        cdef int n
        if not isinstance(x, Rational):
            x = Rational(x)
        n = self.cmp(x)
        if op == 0:
            return bool(n < 0)
        elif op == 1:
            return bool(n <= 0)
        elif op == 2:
            return bool(n == 0)
        elif op == 3:
            return bool(n != 0)
        elif op == 4:
            return bool(n > 0)
        elif op == 5:
            return bool(n >= 0)
            
    def copy(self):
        cdef Rational z
        z = Rational()
        mpq_set(z.value, self.value)
        return z

    def  __dealloc__(self):
        mpq_clear(self.value)
        
    def __repr__(self):
        cdef char *s
        s = mpq_get_str(NULL, 10, self.value)
        t = str(s)
        free(s)
        return t
        
    def __float__(self):
        return mpq_get_d(self.value)
        
    def __hash__(self):
        return hash(mpq_get_d(self.value))
    
    def set_si(self, signed long int n):
        mpq_set_si(self.value, n, 1)

    def set_str(self, s, base=10):
        valid = mpq_set_str(self.value, s, base)
        if valid != 0:
            raise ValueError, "invalid literal:" + s

    def add(Rational self, Rational other):
        cdef Rational x
        x = Rational()
        mpq_add(x.value, self.value, other.value)
        return x
        
    def __add__(x, y):
        if isinstance(x, Rational) and isinstance(y, Rational):
            return x.add(y)
        return coerce.bin_op(x, y, operator.add)

    def sub(Rational self, Rational other):
        cdef Rational x
        x = Rational()
        mpq_sub(x.value, self.value, other.value)
        return x
        
    def __sub__(x, y):
        if isinstance(x, Rational) and isinstance(y, Rational):
            return x.sub(y)
        return coerce.bin_op(x, y, operator.sub)
        
    def mul(Rational self, Rational other):
        cdef Rational x
        x = Rational()
        mpq_mul(x.value, self.value, other.value)
        return x
    
    def __mul__(x, y):
        if isinstance(x, Rational) and isinstance(y, Rational):
            return x.mul(y)
        return coerce.bin_op(x, y, operator.mul)
    
    def div(Rational self, Rational other):
        cdef Rational x
        x = Rational()
        mpq_div(x.value, self.value, other.value)
        return x
        
    def __div__(x, y):
        if isinstance(x, Rational) and isinstance(y, Rational):
            return x.div(y)
        return coerce.bin_op(x, y, operator.div)
    
    def __invert__(self):
        cdef Rational x
        x = Rational()
        mpq_inv(x.value, self.value)
        return x

    def __pow__(Rational self, long int n, dummy):
        cdef Rational x
        x = Rational()
        cdef mpz_t num, den
        mpz_init(num)
        mpz_init(den)
        mpz_pow_ui(num, mpq_numref(self.value), n)
        mpz_pow_ui(den, mpq_denref(self.value), n)
        mpq_set_num(x.value, num)
        mpq_set_den(x.value, den)
        mpz_clear(num)
        mpz_clear(den)        
        return x

    def __pos__(self):
        return self

    def __neg__(self):
        cdef Rational x
        x = Rational()
        mpq_neg(x.value, self.value)
        return x

    def __abs__(self):
        cdef Rational x
        x = Rational()
        mpq_abs(x.value, self.value)
        return x

    def mod_ui(Rational self, unsigned long int n):
        cdef unsigned int num, den, a
        cdef Rational x
        num = mpz_fdiv_ui(mpq_numref(self.value), n)
        den = mpz_fdiv_ui(mpq_denref(self.value), n)
        return int((num * inverse_mod_int(den, n)) % n)

    def parent(self):
        return rational_field.Q

    def numer(self):
        n = integer.Integer()
        set_mpz(<void*> n, mpq_numref(self.value))
        return n

    def numerator(self):
        n = integer.Integer()
        set_mpz(<void*> n, mpq_numref(self.value))
        return n
        
    def __int__(self):
        if mpz_cmp_si(mpq_denref(self.value),1) != 0:
            raise ValueError, "cannot convert %s to an int"%self
        return int(self.numerator())
        
    def __long__(self):
        return int(self)
        
    def denom(self):
        d = integer.Integer()
        set_mpz(<void*> d, mpq_denref(self.value))
        return d

    def denominator(self):
        d = integer.Integer()
        set_mpz(<void*> d, mpq_denref(self.value))
        return d

    def factor(self):
        return rational_field.factor(self)
        
    def height(self):
        """
        The max absolute value of the numerator and denominator of self,
        as an rings.integer.Integer.
        """
        x = abs(self.numer())
        if x > self.denom():
            return x
        return self.denom()
        
    def order(self):
        if mpz_cmp_si(mpq_denref(self.value), 1) != 0:
            return 0
        if mpz_cmp_si(mpq_numref(self.value), 1) == 0:
            return 1
        elif mpz_cmp_si(mpq_numref(self.value), -1) == 0:
            return 2
        return 0

def rational_reconstruction(a, m):
    """
    Find the rational reconstruction of a mod m, if it exists.
    INPUT:
        a -- rings.integer.Integer
        m -- rings.integer.Integer
    OUTPUT:
        x -- rings.rational.Rational
    """
    cdef Rational x
    if not isinstance(a, integer.Integer) or not isinstance(m, integer.Integer):
        a = integer.Integer(a)
        m = integer.Integer(m)
    x = Rational()
    mpq_rational_reconstruction(x.value, get_value(<void*> a)[0], get_value(<void *>m)[0])
    return x

# def test(n):
#     cdef mpz_t a, m
#     mpz_init(a); mpz_init(m)
#     mpz_set_str(a, "133173434946179363436783721312585228097", 0)
#     mpz_set_str(m, "10000000000000000000000000000000000000000", 0)
#     cdef int k
#     cdef mpq_t x
#     mpq_init(x)
#     import time
#     t = time.clock()
#     for k from 0 <= k < n:
#         mpq_rational_reconstruction(x, a, m)
#     print time.clock() - t
